/*
 * philosophers/philosophers.c --
 *
 *	A solution for the dining philosophers problem. This is an
 *      arbitrator solution and goes back to the solution outlined in
 *      Tanenbaum's operating systems book.
 */

#define _POSIX_C_SOURCE 200809L

#include <pthread.h>
#include <stdio.h>
#include <time.h>

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

#define PHIL_STATE_THINKING	0x01
#define PHIL_STATE_HUNGRY	0x02
#define PHIL_STATE_EATING	0x03

typedef struct philosopher {
    char *name;
    int state;
    unsigned meals;
    struct philosopher *left;
    struct philosopher *right;
    pthread_t tid;
    pthread_mutex_t lock;
} philosopher_t;

const struct timespec eat_time   = { .tv_sec = 0, .tv_nsec = 4321 };
const struct timespec think_time = { .tv_sec = 0, .tv_nsec = 1234 };

static void
think(philosopher_t *phil)
{
    printf("%s is thinking\n", phil->name);
    nanosleep(&think_time, NULL);
}

static void
eat(philosopher_t *phil)
{
    phil->meals++;
    printf("%s is eating meal #%d\n", phil->name, phil->meals);
    nanosleep(&eat_time, NULL);
}

static void
test(philosopher_t *phil)
{
    if (phil->state == PHIL_STATE_HUNGRY
	&& phil->left->state != PHIL_STATE_EATING
	&& phil->right->state != PHIL_STATE_EATING) {
	phil->state = PHIL_STATE_EATING;
	(void) pthread_mutex_unlock(&phil->lock);
    }
}

static void
take_forks(philosopher_t *phil)
{
    (void) pthread_mutex_lock(&mutex);
    phil->state = PHIL_STATE_HUNGRY;
    printf("%s is hungry\n", phil->name);
    test(phil);
    (void) pthread_mutex_unlock(&mutex);
    (void) pthread_mutex_lock(&phil->lock);
}

static void
put_forks(philosopher_t *phil)
{
    (void) pthread_mutex_lock(&mutex);
    phil->state = PHIL_STATE_THINKING;
    test(phil->left);
    test(phil->right);
    (void) pthread_mutex_unlock(&mutex);
}

static void*
life(void *arg)
{
    philosopher_t *phil = (philosopher_t *) arg;

    for (int i = 0; i < 1234; i++) {
	think(phil);
	take_forks(phil);
	eat(phil);
	put_forks(phil);
    }

    return NULL;
}

int
main(int argc, char *argv[])
{
    philosopher_t phils[argc];

    for (int i = 1; i < argc; i++) {
	phils[i].name = argv[i];
	phils[i].state = PHIL_STATE_THINKING;
        phils[i].meals = 0;
	(void) pthread_mutex_init(&phils[i].lock, NULL);
	(void) pthread_mutex_unlock(&phils[i].lock);
	phils[i].left  = (i == 1)      ? &phils[argc-1] : &phils[i-1];
	phils[i].right = (i == argc-1) ? &phils[1]      : &phils[i+1];
        (void) pthread_create(&phils[i].tid, NULL, life, &phils[i]);
    }

    (void) pthread_exit(NULL);
    return 0;
}
